from typing import List, Union

from swarmstar.swarm.types import SwarmConfig, SwarmOperation
from swarmstar.utils.swarm_utils.blocking_operations.main import blocking
from swarmstar.utils.swarm_utils.failure_operations.main import failure
from swarmstar.utils.swarm_utils.spawn_operations.main import spawn
from swarmstar.utils.swarm_utils.termination_operations.main import terminate


def swarmstar_god(
    swarm: SwarmConfig, swarm_operation: SwarmOperation
) -> Union[List[SwarmOperation], None]:
    """
    This function is the main entry point for the swarmstar library. It takes in a swarm configuration and a swarm operation
    and returns a list of swarm operations that should be executed next.
    """
    operation_mapping = {
        "spawn": spawn,
        "blocking": blocking,
        "terminate": terminate,
        "node_failure": failure,
    }

    if swarm_operation.operation_type in operation_mapping:
        output = operation_mapping[swarm_operation.operation_type](
            swarm, swarm_operation
        )
    else:
        raise ValueError(
            f"Unknown swarm operation type: {swarm_operation.operation_type}"
        )

    if not isinstance(output, list) and output is not None:
        output = [output]
    return output



from __future__ import annotations

import subprocess
from typing import TYPE_CHECKING

import pymongo
from pymongo import MongoClient

if TYPE_CHECKING:
    from swarmstar.swarm.types import SwarmConfig


def check_and_create_database(mongodb_uri: str, db_name: str) -> None:
    """
    Check if a MongoDB database exists, and if not, create it.

    :param mongodb_uri: MongoDB URI for connecting to the instance.
    :param db_name: Name of the database to check and potentially create.
    :raises ValueError: If the database already exists.
    """
    client = MongoClient(mongodb_uri)
    if db_name in client.list_database_names():
        raise ValueError(f"Database {db_name} already exists.")
    else:
        db = client[db_name]
        collection = db["stage"]
        collection.insert_one({"_id": "dummy_id"})
        collection.delete_one({"_id": "dummy_id"})
        print(f"Database {db_name} created successfully.")


def restore_database(dump_path: str, uri: str, db_name: str):
    """
    Restore a MongoDB database dump to a specified database.

    :param dump_path: Path to the directory containing the dump files.
    :param uri: MongoDB URI for connecting to the target instance.
    :param db_name: Name of the database to restore the dump into.
    """
    # Construct the mongorestore command
    restore_command = ["mongorestore", "--uri", uri, "--db", db_name, dump_path]

    try:
        # Execute the mongorestore command
        subprocess.run(restore_command, check=True)
        print("Database restored successfully.")
    except subprocess.CalledProcessError as e:
        raise ValueError(f"Failed to restore database: {e}")


def create_client(uri: str) -> MongoClient:
    try:
        client = MongoClient(uri)
        return client
    except Exception as e:
        raise ValueError(f"Failed to create MongoDB client: {str(e)}")


def create_collection(swarm: SwarmConfig, category: str) -> None:
    uri = swarm.mongodb_uri
    db_name = swarm.mongodb_db_name
    client = create_client(uri)
    db = client[db_name]
    collection = db[category]
    collection.create_index([("key", pymongo.ASCENDING)], unique=True)


def add_kv(swarm: SwarmConfig, category: str, key: str, value: dict) -> None:
    try:
        db_name = swarm.mongodb_db_name
        collection_name = category
        uri = swarm.mongodb_uri
        client = create_client(uri)
        db = client[db_name]
        if collection_name not in db.list_collection_names():
            create_collection(swarm, collection_name)
        collection = db[collection_name]
        document = {"key": key, **value}
        collection.insert_one(document)
    except pymongo.errors.DuplicateKeyError as e:
        update_kv(swarm, category, key, value)
    except Exception as e:
        raise ValueError(f"Failed to add to MongoDB collection: {str(e)}")


def get_kv(swarm: SwarmConfig, category: str, key: str) -> dict:
    try:
        uri = swarm.mongodb_uri
        db_name = swarm.mongodb_db_name
        collection_name = category
        client = create_client(uri)
        db = client[db_name]
        if collection_name not in db.list_collection_names():
            create_collection(swarm, collection_name)
        collection = db[collection_name]
        result = collection.find_one({"key": key})
        if result is None:
            raise ValueError(f"Key {key} not found in MongoDB collection.")
        result.pop("_id")
        result.pop("key")
        return result
    except Exception as e:
        raise ValueError(f"Failed to get from MongoDB collection: {str(e)}")


def delete_kv(swarm: SwarmConfig, category: str, key: str) -> None:
    try:
        uri = swarm.mongodb_uri
        db_name = swarm.mongodb_db_name
        collection_name = category
        client = create_client(uri)
        db = client[db_name]
        if collection_name not in db.list_collection_names():
            create_collection(swarm, collection_name)
        collection = db[collection_name]
        result = collection.delete_one({"key": key})
        if result.deleted_count == 0:
            raise ValueError(f"Key {key} not found in MongoDB collection.")
    except Exception as e:
        raise ValueError(f"Failed to delete from MongoDB collection: {str(e)}")


def update_kv(swarm: SwarmConfig, category: str, key: str, value: dict) -> None:
    try:
        uri = swarm.mongodb_uri
        db_name = swarm.mongodb_db_name
        collection_name = category
        client = create_client(uri)
        db = client[db_name]
        collection = db[collection_name]
        if collection_name not in db.list_collection_names():
            create_collection(swarm, collection_name)
        result = collection.update_one({"key": key}, {"$set": value})
        if result.matched_count == 0:
            raise ValueError(f"Key {key} not found in MongoDB collection.")
    except Exception as e:
        raise ValueError(f"Failed to update MongoDB collection: {str(e)}")



"""
This module contains functions to load resources from within the package.
"""
import json
from importlib import resources
from typing import Any, BinaryIO

from swarmstar.swarm.types.swarm_config import SwarmConfig
from swarmstar.utils.data.kv_operations.main import get_kv


def get_internal_action_metadata(swarm: SwarmConfig, action_id: str) -> dict:
    return get_internal_mongodb_value(swarm, "action_space", action_id)


def get_internal_memory_metadata(swarm: SwarmConfig, memory_id: str) -> dict:
    return get_internal_mongodb_value(swarm, "memory_space", memory_id)


def get_internal_util_metadata(swarm: SwarmConfig, util_id: str) -> dict:
    return get_internal_mongodb_value(swarm, "util_space", util_id)


def get_internal_mongodb_value(swarm: SwarmConfig, category: str, key: str) -> dict:
    get_kv(swarm, category, key)


def get_json_data(package: str, resource_name: str) -> Any:
    with resources.open_text(package, resource_name) as file:
        return json.load(file)


def get_binary_data(package: str, resource_name: str) -> bytes:
    with resources.open_binary(package, resource_name) as file:
        return file.read()


def get_binary_file(package: str, resource_name: str) -> BinaryIO:
    return resources.open_binary(package, resource_name)
